package org.imogene.gwt.widgets.client.dynaTable;

import java.util.HashSet;
import java.util.List;
import java.util.Set;

import org.imogene.gwt.widgets.client.dynaTable.DynaTableDataProvider.RowDataAcceptor;

import com.google.gwt.core.client.GWT;
import com.google.gwt.event.dom.client.ClickEvent;
import com.google.gwt.event.dom.client.ClickHandler;
import com.google.gwt.event.dom.client.KeyCodes;
import com.google.gwt.event.dom.client.KeyPressEvent;
import com.google.gwt.event.dom.client.KeyPressHandler;
import com.google.gwt.user.client.ui.CheckBox;
import com.google.gwt.user.client.ui.Composite;
import com.google.gwt.user.client.ui.Grid;
import com.google.gwt.user.client.ui.HTML;
import com.google.gwt.user.client.ui.HasHorizontalAlignment;
import com.google.gwt.user.client.ui.HasVerticalAlignment;
import com.google.gwt.user.client.ui.HorizontalPanel;
import com.google.gwt.user.client.ui.Image;
import com.google.gwt.user.client.ui.Label;
import com.google.gwt.user.client.ui.PushButton;
import com.google.gwt.user.client.ui.TextBox;
import com.google.gwt.user.client.ui.VerticalPanel;
import com.google.gwt.user.client.ui.Widget;
import com.google.gwt.user.client.ui.HTMLTable.Cell;


/**
 * Dynamic table.
 * 
 * @author MEDES-IMPS
 */
public class DynaTableWidget extends Composite {
	
	private final VerticalPanel panel = new VerticalPanel();
	private final Grid grid = new Grid();
	private final NavBar navbar = new NavBar();
	
	private final RowDataAcceptor acceptor = new RowDataAcceptorImpl();	
	private final DynaTableDataProvider provider;
	private int startRow = 0;
	private HashSet<String> selectedRow = new HashSet<String>();
	private boolean isCheckable=false;
	
	// property to sort rows
	private Integer sortColumn = null;
	
	// for quick search
	private TextBox valueFilter;	
	private PushButton filterButton;
	private boolean checkBoxesVisible=true;
	
	
	
	public boolean isCheckBoxesVisible() {
		return checkBoxesVisible;
	}

	public void setCheckBoxesVisible(boolean checkBoxesVisible) {
		this.checkBoxesVisible = checkBoxesVisible;
	}

	/**
	 * construct a dynamic table widget
	 * 
	 * @param provider the data provider for this table
	 * @param columns column names
	 * @param columnStyles columns styles
	 * @param rowCount number of row that this table contains
	 */
	public DynaTableWidget(DynaTableDataProvider provider, ColumnHeader[] columns,
			String[] columnStyles, int rowCount) {

		if (columns.length == 0) {
			throw new IllegalArgumentException(
					"DynaTableWidget: expecting a positive number of columns");
		}

		if (columnStyles != null && columns.length != columnStyles.length) {
			throw new IllegalArgumentException(
					"DynaTableWidget: expecting as many styles as columns");
		}

		this.provider = provider;
		
		initWidget(panel);
		panel.add(grid);
		panel.add(navbar);	
		panel.add(createFilterBar());

		initTable(columns, columnStyles, rowCount);
		
		setStyleName("DynaTable-DynaTableWidget");
		grid.setStyleName("table");
	}

	/**
	 * Initialize the table
	 * 
	 * @param columns columns names
	 * @param columnStyles columns styles
	 * @param rowCount number of row
	 */
	private void initTable(ColumnHeader[] columns, String[] columnStyles, int rowCount) {
		/*
		 * Set up the header row. It's one greater than the number of visible
		 * rows
		 */
		grid.resize(rowCount + 1, columns.length+1);
		
		/* set column header : text and style */
		for (int i = 0; i < columns.length; i++) {
						
			columns[i].setStyleName("DynaTable-isNotSortProperty");
			grid.setWidget(0, i+1, columns[i]);
			
			if (columnStyles != null) {
				grid.getCellFormatter().setStyleName(0, i+1, columnStyles[i] + " header");				
			}
		}
		/* set row background color */
		for(int j=1; j< rowCount+1; j++){
			if(j%2==0)
				grid.getRowFormatter().setStyleName(j, "row_1");
			else
				grid.getRowFormatter().setStyleName(j, "row_2");
		}
		

	}

	/**
	 * Set the table status text
	 * 
	 * @param text status text
	 */
	public void setStatusText(String text) {
		navbar.status.setText(text);
	}
	
	/**
	 * @param property the bean property that should be used for 
	 * sorting the table
	 */
	public void setSortColumn(int i) {
		
		ColumnHeader header = (ColumnHeader)grid.getWidget(0, i);
		String property = header.getProperty();
		
		if (property!=null) {
			
			//remove style name from previous sort column
			if (sortColumn!=null) {
				grid.getWidget(0, sortColumn).removeStyleName("DynaTable-isSortProperty");	
				grid.getWidget(0, sortColumn).addStyleName("DynaTable-isNotSortProperty");
			}
						
			//set index of the column to be sorted
			this.sortColumn= i;
			
			// set the sort order to the opposite of the current sort order
			header.setAsc(!header.isAsc());
			
			//set style name for the selected sort column		
			header.removeStyleName("DynaTable-isNotSortProperty");	
			header.addStyleName("DynaTable-isSortProperty");
			
			// set the start row to 0
			startRow= 0;						
		}		
	}

	/**
	 * Clear the table status text
	 */
	public void clearStatusText() {
		navbar.status.setHTML("&nbsp;");
	}

	/**
	 * Refresh the table content
	 */
	public void refresh() {
		
		// Disable buttons temporarily to stop the user from running off the end.
		navbar.gotoFirst.setEnabled(false);
		navbar.gotoPrev.setEnabled(false);
		navbar.gotoNext.setEnabled(false);
		navbar.gotoLast.setEnabled(false);
		navbar.pageNb.setReadOnly(true);
		setStatusText("");	
		navbar.isWaiting();
		
		if (sortColumn==null)
			provider.updateRowData(startRow, grid.getRowCount() - 1, acceptor);
		else {
			ColumnHeader header = (ColumnHeader)grid.getWidget(0, sortColumn);
			provider.updateRowData(startRow, grid.getRowCount() - 1, acceptor, header.getProperty(), header.isAsc());
		}
			
	}
	
	/**
	 * Return a set of rows that are checked.
	 * The set contains String that represents
	 * Id of entity represented by the row.
	 * 
	 * @return a set of row/entity IDs
	 */
	public Set<String> getSelectedRowIds(){
		return selectedRow;
	}
	
	/**
	 * Unselect the row presenting the entity 
	 * identified by the specified ID
	 * 
	 * @param id The entity ID
	 */
	public void unSelect(String id){
		selectedRow.remove(id);
	}

	/**
	 * Set the number of row for this table.
	 * 
	 * @param rows number of rows
	 */
	public void setRowCount(int rows) {
		grid.resizeRows(rows);
	}

	/**
	 * Return the number of row that 
	 * are dedicated to display data.
	 * 
	 * @return number of data row
	 */
	private int getDataRowCount() {
		return grid.getRowCount() - 1;
	}

	/**
	 * Add listener to this table.
	 * 
	 * @param listener listener to add
	 */
	public void addClickHandler(ClickHandler listener) {
		grid.addClickHandler(listener);
	}
	
	/**
	 * Get the source cell for the specified event
	 * @param event the event
	 * @return The source Cell or null
	 */
	public Cell getCellForEvent(ClickEvent event){
		return grid.getCellForEvent(event);
	}
	
	/**
	 * Return all ids of row displayed by the table.
	 * 
	 * @return list a row ids as <code>java.lang.String</code>
	 */
	public List<String> getIds(){
		return provider.getRowDataIds();		
	}
	
	
	/**
	 * Are the table rows checkable
	 * 
	 * @return true if rows are checkable
	 */
	public boolean isCheckable() {
		return isCheckable;
	}

	/**
	 * Set if the table is checkable.
	 * 
	 * @param isCheckable true if the rows are checkable
	 */
	public void setCheckable(boolean isCheckable) {
		this.isCheckable = isCheckable;
	}
	
	
	/**
	 * Create the Widget that permits to enter data for the full text filter.
	 * @return the widget
	 */
	private Widget createFilterBar() {
		
		HorizontalPanel bar = new HorizontalPanel();			
		bar.setStyleName("DynaTable-navbar");	
		
		/* text box to fix the filter */
		HorizontalPanel filterPanel = new HorizontalPanel();
		filterPanel.setStyleName("DynaTable-status");
		valueFilter = new TextBox();
		valueFilter.addKeyPressHandler(new KeyPressHandler() {				
			public void onKeyPress(KeyPressEvent event) {
				
				if (event.getCharCode() == (char) KeyCodes.KEY_ENTER) {
					provider.fullTextSearch(valueFilter.getText());
					startRow=0;
					refresh();										
				}	
			}
		});			
		valueFilter.setReadOnly(false);		
		valueFilter.setVisibleLength(20);
		valueFilter.setStyleName("DynaTable-status-textbox");		
		filterPanel.add(valueFilter);
		filterPanel.setCellVerticalAlignment(valueFilter, HasVerticalAlignment.ALIGN_MIDDLE);
		
		/* button filter */
		HorizontalPanel buttonPanel = new HorizontalPanel();
		buttonPanel.setSpacing(2);
		filterButton = new PushButton(DynaTableNLS.constants().quick_search());
		filterButton.setStyleName("DynaTable-button-search");
		filterButton.addClickHandler(new ClickHandler(){
			public void onClick(ClickEvent w) {
				provider.fullTextSearch(valueFilter.getText());
				startRow=0;
				refresh();
			}
		});
		buttonPanel.add(filterButton);
		buttonPanel.setCellVerticalAlignment(filterButton, HasVerticalAlignment.ALIGN_MIDDLE);
		
		bar.add(filterPanel);
		bar.add(buttonPanel);
				
		bar.setCellHorizontalAlignment(filterPanel, HasHorizontalAlignment.ALIGN_LEFT);
		bar.setCellVerticalAlignment(filterPanel, HasVerticalAlignment.ALIGN_MIDDLE);		
		bar.setCellHorizontalAlignment(buttonPanel, HasHorizontalAlignment.ALIGN_LEFT);
		bar.setCellVerticalAlignment(buttonPanel, HasVerticalAlignment.ALIGN_MIDDLE);
		bar.setCellWidth(buttonPanel, "100%");
				
		return bar;
	}
	
	
	/* ********************************* *
	 *         PRIVATE CLASSSES
	 * ********************************* */
	
	/**
	 * Navigation bar
	 * 
	 * @author MEDES-IMPS
	 */
	private class NavBar extends Composite implements ClickHandler, KeyPressHandler {

		public final HorizontalPanel bar = new HorizontalPanel();
		public final PushButton gotoFirst = new PushButton("<<", this);
		public final PushButton gotoNext = new PushButton(">", this);
		public final PushButton gotoPrev = new PushButton("<", this);
		public final PushButton gotoLast = new PushButton(">>", this);
		public final TextBox pageNb = new TextBox();
		public final Label pageTotalNb = new Label();
		public final Label rowTotalNb = new Label();		
		public final HTML status = new HTML();		
		public final HorizontalPanel rowStatus = new HorizontalPanel();
		
		/**
		 * create table navigation bar
		 */
		public NavBar() {
			initWidget(bar);
			bar.setStyleName("DynaTable-navbar");
			status.setStyleName("DynaTable-status");
			
			// Page Panel
			HorizontalPanel pagePanel = new HorizontalPanel();
			pagePanel.setStyleName("DynaTable-status");	
			
			// Reward Buttons
			HorizontalPanel buttonsFirst = new HorizontalPanel();
			buttonsFirst.setSpacing(2);
			gotoFirst.setStyleName("DynaTable-navbar-button");
			gotoPrev.setStyleName("DynaTable-navbar-button");
			buttonsFirst.add(gotoFirst);
			buttonsFirst.add(gotoPrev);
			
			// Page status
			HorizontalPanel pagesStatus = new HorizontalPanel();	

			Label pageLabel = new Label(DynaTableNLS.constants().navbar_page());
			pageLabel.setStyleName("DynaTable-status-text");
			pagesStatus.add(pageLabel);		
			pageNb.setText("1");
			pageNb.setStyleName("DynaTable-status-textbox");
			pageNb.setVisibleLength(4);
			pageNb.addKeyPressHandler(this);
			pagesStatus.add(pageNb);				
			Label pageOf = new Label(DynaTableNLS.constants().navbar_of());
			pageOf.setStyleName("DynaTable-status-text");
			pagesStatus.add(pageOf);			
			pageTotalNb.setText("1");
			pageTotalNb.setStyleName("DynaTable-status-text");
			pagesStatus.add(pageTotalNb);
									
			// Forward Buttons
			HorizontalPanel buttonsLast = new HorizontalPanel();
			buttonsLast.setSpacing(2);
			gotoNext.setStyleName("DynaTable-navbar-button");
			gotoLast.setStyleName("DynaTable-navbar-button");
			buttonsLast.add(gotoNext);
			buttonsLast.add(gotoLast);
			
			// Row count status
			rowStatus.addStyleName("DynaTable-status");
			Label rowLabel = new Label(DynaTableNLS.constants().navbar_items()+ " ");
			rowLabel.setStyleName("DynaTable-status-text");
			rowStatus.add(rowLabel);
			status.setStyleName("DynaTable-status-text");
			rowStatus.add(status);
			Label rowOf = new Label(DynaTableNLS.constants().navbar_of());
			rowOf.setStyleName("DynaTable-status-text");
			rowStatus.add(rowOf);
			rowTotalNb.setStyleName("DynaTable-status-text");
			rowTotalNb.setText("0");
			rowStatus.add(rowTotalNb);			
									
			pagePanel.add(buttonsFirst);
			pagePanel.add(pagesStatus);
			pagePanel.add(buttonsLast);
			
			bar.add(pagePanel);
			bar.add(rowStatus);
			
			pagePanel.setCellVerticalAlignment(buttonsFirst, HasVerticalAlignment.ALIGN_MIDDLE);
			pagePanel.setCellVerticalAlignment(buttonsLast, HasVerticalAlignment.ALIGN_MIDDLE);
			pagePanel.setCellVerticalAlignment(pagesStatus, HasVerticalAlignment.ALIGN_MIDDLE);			
			buttonsFirst.setCellVerticalAlignment(gotoFirst, HasVerticalAlignment.ALIGN_MIDDLE);
			buttonsFirst.setCellVerticalAlignment(gotoPrev, HasVerticalAlignment.ALIGN_MIDDLE);		
			pagesStatus.setCellVerticalAlignment(pageLabel, HasVerticalAlignment.ALIGN_MIDDLE);
			pagesStatus.setCellVerticalAlignment(pageNb, HasVerticalAlignment.ALIGN_MIDDLE);
			pagesStatus.setCellVerticalAlignment(pageOf, HasVerticalAlignment.ALIGN_MIDDLE);
			pagesStatus.setCellVerticalAlignment(pageTotalNb, HasVerticalAlignment.ALIGN_MIDDLE);		
			buttonsLast.setCellVerticalAlignment(gotoNext, HasVerticalAlignment.ALIGN_MIDDLE);
			buttonsLast.setCellVerticalAlignment(gotoLast, HasVerticalAlignment.ALIGN_MIDDLE);
			
			bar.setCellHorizontalAlignment(pagePanel, HasHorizontalAlignment.ALIGN_LEFT);
			bar.setCellVerticalAlignment(pagePanel, HasVerticalAlignment.ALIGN_MIDDLE);		
			bar.setCellHorizontalAlignment(rowStatus, HasHorizontalAlignment.ALIGN_RIGHT);
			bar.setCellVerticalAlignment(rowStatus, HasVerticalAlignment.ALIGN_MIDDLE);

			// Initialize prev & first button to disabled.
			gotoPrev.setEnabled(false);
			gotoFirst.setEnabled(false);
		}

		/*
		 * (non-Javadoc)
		 * @see com.google.gwt.event.dom.client.ClickHandler#onClick(com.google.gwt.event.dom.client.ClickEvent)
		 */
		public void onClick(ClickEvent event) {
			if (event.getSource() == gotoNext) {
				startRow += getDataRowCount();
				refresh();
			} else if (event.getSource() == gotoPrev) {
				startRow -= getDataRowCount();
				if (startRow < 0) {
					startRow = 0;
				}
				refresh();
			} else if (event.getSource() == gotoFirst) {
				startRow = 0;
				refresh();
			} else if (event.getSource() == gotoLast) {
				int rowNb = Integer.valueOf(rowTotalNb.getText());
				int pageNb = Integer.valueOf(pageTotalNb.getText());
				int rowPerPage = getDataRowCount();
				startRow = rowNb - (rowPerPage - ((pageNb*rowPerPage)-rowNb));
				refresh();
			}
		}
				
		/*
		 * (non-Javadoc)
		 * @see com.google.gwt.event.dom.client.KeyPressHandler#onKeyPress(com.google.gwt.event.dom.client.KeyPressEvent)
		 */
		public void onKeyPress(KeyPressEvent event) {
			
			if (event.getSource() == pageNb) {
				
				if ((!Character.isDigit(event.getCharCode())) 
						//&& (keyCode != (char) KEY_TAB)
						&& (event.getCharCode() != (char) KeyCodes.KEY_BACKSPACE)
						&& (event.getCharCode() != (char) KeyCodes.KEY_DELETE) 
						&& (event.getCharCode() != (char) KeyCodes.KEY_ENTER) 
						//&& (keyCode != (char) KEY_HOME) 
						&& (event.getCharCode() != (char) KeyCodes.KEY_END)
						&& (event.getCharCode() != (char) KeyCodes.KEY_LEFT) 
						//&& (keyCode != (char) KEY_UP)
						//&& (keyCode != (char) KEY_DOWN) 
						&& (event.getCharCode() != (char) KeyCodes.KEY_RIGHT)) {

					((TextBox)event.getSource()).cancelKey();					
				}
				else {
					if (event.getCharCode() == (char) KeyCodes.KEY_ENTER) {
						
						int pageNum = Integer.valueOf(pageNb.getText());
						int pageNb = Integer.valueOf(pageTotalNb.getText());							
						if (pageNum<=pageNb) {							
							int rowPerPage = getDataRowCount();							
							int newStartRow = ((pageNum-1)*rowPerPage);
							if (newStartRow!=startRow){
								startRow = newStartRow;						
								refresh();									
							}						
						}					
					}					
				}
			}			
		}		
		
		public void isWaiting() {
			Image img = new Image(GWT.getModuleBaseURL()+ "images/loading.gif");
			rowStatus.setCellVerticalAlignment(img, HasVerticalAlignment.ALIGN_MIDDLE);
			rowStatus.insert(img, 2);
		}
		
		public void stopWaiting() {
			rowStatus.remove(2);
		}

	}

	/**
	 * Populate table with widget adapted 
	 * to the data to display.
	 * 
	 * @author MEDES-IMPS
	 */
	private class RowDataAcceptorImpl implements RowDataAcceptor {
		
		private String[] entitiesIds;
		
		/**
		 * @see RowDataAcceptor#accept(int, String[][])
		 */
		public void accept(int startRow, String[][] data, int totalNbOfRows) {
			
			Label[][] widgetData = null;
			
			if (data!=null) {
				widgetData = new Label[data.length][];

				for (int i=0; i<data.length; i++) {					
					String[] row = data[i];					
					widgetData[i] = new Label[row.length];					
					for (int j=0; j<row.length; j++) {
						String item = row[j];
						widgetData[i][j].setText(item);					
					}				
				}								
			}			
			accept(startRow, widgetData, totalNbOfRows);
		}

		
		/**
		 * Accept widget data for checkable table
		 * 
		 * @param startRow start at this row
		 * @param data data as widget to display 
		 * @param entitiesID data associated entities ids
		 */
		public void accept(int startRow, Widget[][] data, String[] entitiesID, int totalNbOfRows){
			this.entitiesIds = entitiesID;			
			isCheckable = true;
			accept(startRow, data, true, totalNbOfRows);			
		}
		
		
		public void accept(int startRow, Widget[][] data, int totalNbOfRows) {			
			isCheckable = false;
			accept(startRow, data, false, totalNbOfRows);
		}
		
		/**
		 * accept widget data for not checkable table
		 * 
		 * @param startRow start at this index
		 * @param data as widget to display 
		 */
		protected void accept(int startRow, Widget[][] data, boolean checkable, int totalNbOfRows) {

			int destRowCount = getDataRowCount();
			int destColCount = grid.getCellCount(0);

			int srcRowIndex = 0;
			int srcRowCount = data.length;
			int destRowIndex = 1; // skip navbar row

			if(checkable){
				// clear table
				//clearRows();		
				for (; srcRowIndex < srcRowCount; ++srcRowIndex, ++destRowIndex) {
					Widget[] srcRowData = data[srcRowIndex];
					boolean check = false;
					if(selectedRow.contains(entitiesIds[destRowIndex-1])){
						check=true;
					}
					
					if(checkBoxesVisible)
						grid.setWidget(destRowIndex, 0, createCheckBox(entitiesIds[destRowIndex-1],check));
					for (int srcColIndex = 1; srcColIndex < destColCount; ++srcColIndex) {
						Widget widget = srcRowData[srcColIndex-1];
						grid.clearCell(destRowIndex, srcColIndex);
						grid.setWidget(destRowIndex, srcColIndex, widget);
					}
				}
				grid.getColumnFormatter().setWidth(0, "28px");
			}
			else{
				for (; srcRowIndex < srcRowCount; ++srcRowIndex, ++destRowIndex) {
					Widget[] srcRowData = data[srcRowIndex];					
					for (int srcColIndex = 1; srcColIndex < destColCount; ++srcColIndex) {
						Widget widget = srcRowData[srcColIndex-1];
						grid.clearCell(destRowIndex, srcColIndex);
						grid.setWidget(destRowIndex, srcColIndex, widget);
					}
				}
			}

			/* Clear remaining table rows. */			
			for (; destRowIndex < destRowCount + 1; ++destRowIndex) {
				for (int destColIndex = 0; destColIndex < destColCount; ++destColIndex) {
					grid.clearCell(destRowIndex, destColIndex);
				}
			}

			/* Update navbar information */
			int totalNbOfPages = 0;			
			if ((totalNbOfRows % destRowCount)==0)
				totalNbOfPages = totalNbOfRows / destRowCount;
			else
				totalNbOfPages = totalNbOfRows / destRowCount + 1;	
			
			int currentPageNb = 0;			
			if ((startRow % destRowCount)==0)
				currentPageNb = (startRow / destRowCount) + 1;
			else
				currentPageNb = (startRow / destRowCount) + 2;							
						
			navbar.rowTotalNb.setText(String.valueOf(totalNbOfRows));
			navbar.pageTotalNb.setText(String.valueOf(totalNbOfPages));
			navbar.pageNb.setText(String.valueOf(currentPageNb));	
			navbar.pageNb.setReadOnly(false);

			/* Update the status message. */
			navbar.stopWaiting();
			if (startRow==0 && srcRowCount==0) {
				setStatusText("0 - 0");
			}
			else {
				setStatusText((startRow + 1) + " - " + (startRow + srcRowCount));
			}
			
			/* Synchronize the navigation buttons. */	
			boolean isLastPage = false;
			if(srcRowCount<destRowCount) isLastPage = true;
			navbar.gotoFirst.setEnabled(startRow > 0);
			navbar.gotoPrev.setEnabled(startRow > 0);
			navbar.gotoNext.setEnabled(!isLastPage);
			navbar.gotoLast.setEnabled(!isLastPage);

		}

		/**
		 * Display failure message in the status text area
		 * 
		 * @param caught the error
		 */
		public void failed(Throwable caught) {
			navbar.stopWaiting();
			String msg = DynaTableNLS.constants().application_fail();
			setStatusText(msg);
		}
		
		/**
		 *  Clear all rows in the table.
		 */
		@SuppressWarnings("unused")
		public void clearRows() {
			for(int i=0; i<grid.getRowCount()-1; i++)
			{
				for (int j=0; j<grid.getColumnCount()-1; j++)
				{
					Widget widget = grid.getWidget(i, j);
					if (widget!=null)
					{
						widget.removeFromParent();
						widget=null;
					}	
				}
			}
		}
				
		
		/** 
		 * Create check box that permits to select a row.
		 * 
		 * @param rowId the id of the row
		 * @param check checked if its value is true
		 * @return the check box widget
		 */
		private CheckBox createCheckBox(final String rowId, boolean check){
			CheckBox chkb = new CheckBox();
			chkb.setValue(new Boolean(check));			
			chkb.addClickHandler(new ClickHandler(){

				public void onClick(ClickEvent event) {					
					if(((CheckBox)event.getSource()).getValue()){
						selectedRow.add(rowId);		
					}else{
						selectedRow.remove(rowId);	
					}
				}
				
			});							
			return chkb;			
		}
		

	}



}
